' Basic String List Object
' Richard D. Clark
' Use at your own risk.
' Public Domain
' This program is free software; you can redistribute it and/or modify it
' but WITHOUT ANY WARRANTY; without even the implied warranty of
' MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
' ================================================================

#include once "crt.bi"

'Wrap our object in a custom namespace.
namespace strlist

const strlVersion = "0.1.1"

'Create a NULL value.
#ifndef NULL
#define NULL 0
#endif

'Create a True/False value.
#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE (not FALSE)
#endif

'Needed by the compare function.
dim shared casesense as integer
dim shared sortype as integer

'This is our little stringlist object.
type stringlist
	private:
	_listcount as integer 'The number of items in list.
	_list as zstring ptr ptr 'List of pointers to zstrings.
	_issorted as integer       'Is list sorted.
	_casesense as integer      'True if case sensitive, False if not.
	_sortype as integer        'The last sorting mode.
	_version as string        'The current version.
	declare sub _DestroyList () 'Used to clear the pointer list.
	declare function _AddItemToList(item as string) as integer 'Adds a new item to the list.
	declare static function _QCompare cdecl (e1 as any ptr, e2 as any ptr) as long
	public:
	declare constructor () 'Object constructor.
	declare destructor () 'Used to clean up list, calls _DestroyList. Used when object goes out of scope.
	declare property Version () as string 'Returns the current version.
	declare property Count () as integer 'Returns the number of items in list.
	declare property Sorted () as integer 'Returns true if list is sorted.
	declare property Strings (index as integer, item as string) 'Sets item based on index.
	declare property Strings (index as integer) as string 'Returns item based on index.
	declare property Text (delimiter as string, txt as string) 'Sets stringlist based on text.
	declare property Text (delimiter as string) as string 'Returns string as text delimited by delimiter.
	declare property CaseSensitive (scase as integer) 'Sets the case sensitive value.
	declare property CaseSensitive () as integer 'Returns the case sensitive value.
	declare property SortAscending (sortdir as integer) 'Sets the case sensitive value.
	declare property SortAscending () as integer 'Returns the case sensitive value.
	declare function LoadFromFile (fname as string) as integer 'Loads data from file, loads into list.
	declare function SaveToFile (fname as string) as integer 'Save data to a line-based text file.
	declare function add (item as string) as integer 'Adds an item to the list, returns index value or -1.
	declare function InsertItem (index as integer, item as string) as integer 'Inserts item at index.
	declare function DeleteItem (index as integer) as integer 'Deletes item with index.
	declare function Move (fromindex as integer, toindex as integer) as integer 'Moves item from index to index.
	declare function Find (fkey as string, byref index as integer) as integer 'Returns True if found, False if not. Index contains position.
	declare function IsEqual(sl as stringlist) as integer 'Compares two stringlists and returns True if equal, False if not. 
	declare sub ClearList () 'Resets list. Calls _DestroyList.
	declare sub SwapItems (index1 as integer, index2 as integer) 'Swaps two index items.
	declare sub Sort () 'Sorts list.
	declare sub Copy (byref sl as stringlist) 'Creates a new copy of stringlist. 
end type

'Object constructor, called when created.
constructor stringlist ()
	'Set the default values.
	_issorted = FALSE
	_list = NULL
	_listcount = 0
	_casesense = FALSE
	casesense = FALSE
	_sortype = TRUE
	sortype = TRUE
	_version = strlVersion
end constructor

'This will be called when object goes out of scope. Needed to clean up.
destructor stringlist ()
	_DestroyList
end destructor

'Returns the current version.
property stringlist.Version () as string
	return _version
end property

'Returns the current number of items in list.
property stringlist.Count () as integer
	return _listcount
end property

'Returns true if list is sorted.
property stringlist.Sorted () as integer 
	return _issorted
end property

'Sets the comparison value.
property stringlist.CaseSensitive (scase as integer)
	'Validate the value.
	if scase = TRUE then 
		_casesense = TRUE
	else
		_casesense = FALSE
	end if
	casesense = _casesense
end property

'Returns the comparison value.
property stringlist.CaseSensitive () as integer 
	return _casesense
end property

'Sets the sorting order.
property stringlist.SortAscending (sortdir as integer)
	'Validate the value.
	if sortdir = TRUE then 
		_sortype = TRUE
	else
		_sortype = FALSE
	end if
	sortype = _sortype
end property

'Returns the comparison value.
property stringlist.SortAscending () as integer 
	return _sortype
end property

'Gets an item from the list.
property stringlist.Strings (index as integer) as string
	dim ret as string = ""
	
	'Make sure we have a valid list.
	if _list then
		'Make sure index is within bounds.
		if (index >= 0) and (index <= _listcount - 1) then
			'Make sure we have a valid pointer.
			if _list[index] then
				ret = *_list[index]
			else
				ret = ""
			end if
		endif
	end if
	
	return ret
end property

'Replaces a string item in the list.
property stringlist.Strings(index as integer, item as string)
	dim as integer itemlen
	
	'Make sure we have a valid list.
	if _list then
		'Make sure index is within bounds.
		if (index >= 0) and (index <= _listcount - 1) then
			itemlen = len(item)
			'Make sure we have a string.
			if itemlen > 0 then
				'Adjust the size of the zstring.
				_list[index] = reallocate (_list[index], itemlen + 1)
				'Make sure we have a valid pointer.
				if _list[index] then
					'Set the list item text.
					*_list[index] = item
				else
					'Some error so clear the item.
					deallocate _list[index]
					_list[index] = NULL
				end if
			else
				'If emptyy string deallocate and set to null.
				deallocate _list[index]
				_list[index] = NULL
			end if
		endif
	'Reset the sorted flag.
	_issorted = FALSE
	end if
	
end property

'Sets stringlist based on item text. Parsed using delimiter.
property stringlist.Text (delimiter as string, txt as string)
	dim as string tx, tmpt
	dim as integer idx, ld
	
	'Get length of delimiter.
	ld = len(delimiter)
	'Get copy of text.
	tx = txt
	'Make sure we have a a delimiter.
	if (ld > 0) and (len(tx) > 0) then
		'Clear current list.
		_DestroyList
		do 
			'Extract each record in the text and add to the list.
			idx = instr(tx, delimiter)
			tmpt = left(tx, idx - 1)
			add(tmpt)
			tx = mid(tx, idx + ld)
		loop while len(tx) > 0
	endif 
	
end property 

'Returns string as text delimited by delimiter.
property stringlist.Text (delimiter as string) as string
	dim ret as string = ""
	
	'Make sure have a valid list.
	if _list then
		'Make sure we have a delimiter.
		if len(delimiter) > 0 then
			for i as integer = 0 to _listcount - 1
				ret += Strings(i) & delimiter 
			next
		endif
	end if
	
	return ret
end property 

'This adds an item to the string list.
function stringlist._AddItemToList(item as string) as integer
	dim as integer itemlen
	dim ret as integer = TRUE
	
	'We'll need this for the zstring.
	itemlen = len(item)
	_listcount += 1
	'Add a new item to the list.
	_list = reallocate (_list, _listcount * sizeof (zstring ptr))
	'Make sure we have a valid pointer.
	if _list then
		'Reset the sorted flag.
		_issorted = FALSE
		'Check to see if we have an empty string.
		if itemlen = 0 then
			_list[_listcount - 1] = NULL
		else
			'Create a new zstring.
			_list[_listcount - 1] = allocate(itemlen + 1) 'Add a space for the null terminator.
			if _list[_listcount - 1] then
				'Set the list item text.
				*_list[_listcount - 1] = item
			else
				'Set it to null.
				_list[_listcount - 1] = NULL
			endif
		end if
	else
		ret = FALSE
	endif
	
	return ret
end function

'Adds an item to the list. Will create list if not created. Returns index of item if successful, or
'-1 if not successful.
function stringlist.Add (item as string) as integer
	if _AddItemToList(item) = TRUE then
		return _listcount - 1
	else
		return -1
	endif
	
end function
	 
'Loads a list of items from a file. Clears current list, if any.
function stringlist.LoadFromFile (fname as string) as integer
	dim ret as integer = TRUE
	dim instring as string
	
	'Reset the current state of the object.
	_DestroyList
	'Make sure we get a file name.
	if len(fname) > 0 then
		'Get a file handle.
		dim fh as integer = freefile
		'make sure we can open the file.
		if open (fname for input as #fh) = 0 then
			do while not eof(fh)
				'Load up each item in the file.
				line input #fh, instring
				if len(instring) > 0 then
					'Add the item to the list.
					if _AddItemToList(instring) = FALSE then
						'Set the return value.
						ret = FALSE
						'Can't go on since we have some memory problem.
						exit do
					endif 
				endif
			loop
			'Close file.
			close #fh
		else
			ret = FALSE
		endif
	else
		ret = FALSE
	endif
	'If ret = false, clean up any partial list that may have been created.
	if ret = FALSE then
		_DestroyList
	endif
	'Return the result. True = list created, False = some error.
	return ret
end function

'Save the contexts of the stringlist to a line-based file.
function stringlist.SaveToFile (fname as string) as integer
	dim ret as integer = TRUE
	dim as integer fh = freefile
	dim outstr as string
	
	'Make sure we have a valid list.
	if _list then
		'Make sure we have a file name.
		if len(fname) > 0 then
			'Open file for output.
			if open (fname for output as #fh) = 0 then
				'Get each string and save to file.
				for i as integer = 0 to _listcount - 1
					outstr = Strings(i)
					print #fh, outstr
				next
				'Close the file.
				close #fh
			else
				ret = FALSE
			endif
		else
			ret = FALSE
		endif
	else
		ret = FALSE
	end if
	return ret
		
end function

'Deletes item with index. Returns True if successful, False if not.
function stringlist.DeleteItem (index as integer) as integer
	dim ret as integer = TRUE
	
	'Make sure we have a valid list.
	if _list then
		'Make sure index is within bounds.
		if (index >= 0) and (index <= _listcount - 1) then
			'Deallocate the string from the list.
			deallocate _list[index]
			'Move all the string items down by one.
			for i as integer = index to _listcount - 2
				_list[i] = _list[i + 1]
			next
			'Set last item to NUll so we don't delete string that has moved.
			_list[_listcount - 1] = NULL
			'Delete the last item from the list.
			deallocate _list[_listcount - 1] 
			'Set the new list count.
			_listcount -= 1
			'Make sure we still have some items in list.
			if _listcount > 0 then
				'Resize the list.
				_list = reallocate (_list, _listcount * sizeof (zstring ptr))
			else
				'Clear the list.
				deallocate _list
				_list = NULL
				'Reset the sorted flag.
				_issorted = FALSE
			end if
		else
			ret = FALSE
		endif
	else
		ret = FALSE
	end if
	
	return ret
	
end function

'Inserts item at index. If successful, returns True, else False.
function stringlist.InsertItem(index as integer, item as string) as integer
	dim ret as integer = TRUE
	dim itemlen as integer = len(item)
	
	'Make sure we have a valid list.
	if _list then
		'Make sure index is within bounds.
		if (index >= 0) and (index <= _listcount) then
			'If index is at end of list just add the item.
			if index = _listcount then
				ret = _AddItemToList(item)
			else
				'Increase the list count.
				_listcount += 1
				'Enlarge the list.
				_list = reallocate (_list, _listcount * sizeof (zstring ptr))
				'Move each item up by 1.
				for i as integer = _listcount - 2 to index step -1
					_list[i + 1] = _list[i]
				next
				'Insert new item at index.
				if itemlen > 0 then
					_list[index] = allocate(itemlen + 1) 'Add a space for the null terminator.
					if _list[index] then
						'Set the list item text.
						*_list[index] = item
						'Set the sort flag.
						_issorted = FALSE
					else
						'Set it to null.
						_list[index] = NULL
						'Set the sort flag.
						_issorted = FALSE
					endif
				else
					_list[index] = NULL
				endif
			end if
		else
			ret = FALSE
		end if
	else
		ret = FALSE
	endif
	
	return ret
end function

'Moves item from index to index, adjusting items to make room.
function stringlist.Move (fromindex as integer, toindex as integer) as integer
	dim ret as integer = TRUE
	dim as string item
	
	'Validate the indexes.
	if fromIndex <> toIndex then
		if (fromindex >= 0) and (fromindex <= _listcount - 1) then
			if (toindex >= 0) and (toindex <= _listcount) then
				'Get the string.
				item = Strings(fromindex)
				'Do an insert followed by a delete.
				ret = InsertItem(toindex, item)
				if ret = TRUE then
					ret = DeleteItem(fromindex)
					'Set the sort flag.
					_issorted = FALSE
				endif
			else
				ret = FALSE
			end if
		else
			ret = FALSE
		endif 
	else
		ret = FALSE
	end if
	return ret
end function

'Finds element in list.
function stringlist.Find (fkey as string, byref index as integer) as integer
	dim ret as integer = FALSE
	dim as string item, ffkey = fkey
	
	'Set the default.
	index = -1
	'Make sure we have a valud list.
	if _list then
		'What comparison value is active.
		if _casesense = FALSE then
			ffkey = ucase(ffkey)
		end if
		'If the list is not sorted do a linaear search.
		if (_issorted = FALSE) or (_sortype = FALSE) then
			'Iterate through each element.
			for i as integer = 0 to _listcount - 1
				'If not an empty string get item.
				if _list[i] then
					item = *_list[i]
				else
					item = ""
				endif
				if _casesense = FALSE then
					item = ucase(item)
				end if
				'If found item, set the index and return.
				if ffkey = item then
					index = i
					ret = TRUE
					exit for
				endif 
			next
		else
			'Do a binary search on the list.
			dim as integer imid
			dim as integer first = 0
			dim as integer last = _listcount - 1
			do while  first <= last 
				imid = (first + last) / 2
				'If not an empty string get item.
				if _list[imid] then
					item = *_list[imid]
				else
					item = ""
				endif
				'Set case type.
				if _casesense = FALSE then
					item = ucase(item)
				end if
				'Compare the values.
				if item = ffkey then
					index = imid
					ret = TRUE
					exit do
				elseif item > ffkey then
					last = imid - 1
				else
					first = imid + 1
				end if
			loop
		end if
	end if
	return ret	
end function 

'Compares two stringlists and returns True if equal, False if not.
function stringlist.IsEqual(sl as stringlist) as integer
	dim ret as integer = TRUE
	dim as string t1, t2
	
	if sl.Count = _listcount then
		'Make sure we have some strings in the list.
		if (sl.Count > 0) and (_listcount) > 0 then
			'Iterate through each list and compare strings.
			for i as integer = 0 to _listcount - 1
				'Get the two strings.
				t1 = sl.Strings(i)
				t2 = Strings(i)
				'If not case sensitive.
				if _casesense = FALSE then
					t1 = ucase(t1)
					t2 = ucase(t2)
				endif
				'Compare the strings.
				if t1 <> t2 then
					ret = FALSE
					exit for
				endif
			next
		endif
	else
		ret = FALSE
	endif
	
	return ret
end function 


'Destroys the list, if any.
sub stringlist._DestroyList ()
	'Make sure we have some items in the list.
	if _list then
		for i as integer = _listcount - 1 to 0 step -1
			'Delete each item in list.
			deallocate _list[i]
		next
		'Delete the list.
		deallocate _list
		'Set the list ptr to null.
		_list = NULL
		'Reset the list count.
		_listcount = 0
		'Reset the sorted flag.
		_issorted = FALSE
	endif
end sub


'Clear current list.
sub stringlist.ClearList()
	_DestroyList
end sub

'Swaps two items in the list.
sub stringlist.SwapItems (index1 as integer, index2 as integer)
	
	'Make sure we have a valid list.
	if _list then
		if index1 <> index2 then
			'Make sure indexes is within bounds.
			if (index1 >= 0) and (index1 <= _listcount - 1) then
				if (index2 >= 0) and (index2 <= _listcount - 1) then
					'Don't need to swap if index is same.
					swap _list[index1], _list[index2]
				end if
			end if
		end if
	'Reset the sorted flag.
	_issorted = FALSE
	endif
		
end sub

'Qsort comparison function.
function stringlist._QCompare cdecl (e1 as any ptr, e2 as any ptr) as long
      dim as string el1, el2
      dim ret as integer
        
		'Get the values, must cast to string ptr
		el1 = *(cptr(string ptr, e1))
		el2 = *(cptr(string ptr, e2))
		if casesense = FALSE then
			el1 = ucase(el1)
			el2 = ucase(el2)
		endif			
		'Compare the values
		if sortype = TRUE then
			if el1 < el2 then
		   	ret = -1
			elseif el1 > el2 then
		   	ret = 1
			else
		   	ret = 0
			end if
      else
			if el1 > el2 then
				ret = -1
			elseif el1 < el2 then
	   		ret = 1
			else
   		ret = 0
			end if
		end if
		
		return ret
end function

'Uses CRT qsort function.
sub stringlist.Sort ()
	'Make sure we have a valid list.
	if _list then
		'Build a working array.
		dim slist(0 to _listcount - 1) as string
		'Copy the elements to the array.
		for i as integer = 0 to _listcount - 1
			'Check for empty string.
			if _list[i] then
				slist(i) = *_list[i]
			else
				slist(i) = ""
			endif
		next
		qsort @slist(0), _listcount, sizeof(string), @_QCompare
		'Copy back to string list.
		for i as integer = 0 to _listcount - 1
			Strings(i) = slist(i)
		next
		_issorted = TRUE
	end if
end sub

'Creates a new copy of stringlist.
sub stringlist.Copy (byref sl as stringlist) 
	
	'Copy the properties.
	sl.CaseSensitive = _casesense
	sl.SortAscending = _sortype
	'If we have a list copy it.
	if _list then
		for i as integer = 0 to _listcount - 1
			sl.Add(Strings(i))
		next
	endif
	
end sub

end namespace
